/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * This file is part of SableCC.                             *
 * See the file "LICENSE" for copyright information and the  *
 * terms and conditions for copying, distribution and        *
 * modification of SableCC.                                  *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

package org.sablecc.sablecc;

import java.util.*;
import org.sablecc.sablecc.analysis.*;
import org.sablecc.sablecc.node.*;
import java.io.*;
import org.sablecc.sablecc.Grammar;
import java.util.Vector;
import java.util.Enumeration;

public class GenerateAlternativeCodeForParser extends DepthFirstAdapter
{
  String currentAlt;
  String realcurrentAlt;
  BufferedWriter file;
  private File pkgDir;

  private ResolveTransformIds transformIds;
  private ComputeCGNomenclature CG;
  private ComputeSimpleTermPosition CTP;
  private MacroExpander macros;
  private Map<Node, String> simpleTermTransformMap;
  private List<Node> listSimpleTermTransform;
  private Map<Node, String> simpleTermOrsimpleListTermTypes;

  private int popCount;

  GenerateAlternativeCodeForParser(File pkgDir, String aParsedAltName,
                                   String raParsedAltName,
                                   BufferedWriter file,
                                   ResolveTransformIds transformIds,
                                   ComputeCGNomenclature CG,
                                   ComputeSimpleTermPosition CTP,
                                   Map<Node, String> simpleTermTransformMap,
                                   MacroExpander macros,
                                   List<Node> listSimpleTermTransform,
                                   Map<Node, String> simpleTermOrsimpleListTermTypes)
  {
    this.pkgDir = pkgDir;
    this.file = file;
    currentAlt = aParsedAltName;
    realcurrentAlt = raParsedAltName;
    this.transformIds = transformIds;
    this.CG = CG;
    this.CTP = CTP;
    this.simpleTermTransformMap = simpleTermTransformMap;
    this.macros = macros;
    this.listSimpleTermTransform = listSimpleTermTransform;
    this.simpleTermOrsimpleListTermTypes = simpleTermOrsimpleListTermTypes;

    this.popCount = 0;
  }

  @Override
  public void caseAElem(final AElem node) {
    popCount++;
  }

  @Override
  public void inAAltTransform(AAltTransform node)
  {
    String type_name;
    int position;

    for(PTerm term : node.getTerms())
    {
      if(simpleTermTransformMap.get(term) != null)
      {
        type_name = simpleTermTransformMap.get(term);
      }
      else
      {
        type_name = CG.getAltTransformElemTypes().get(term);
      }

      position = CG.getTermNumbers().get(term);

      try
      {
        if(type_name.startsWith("L"))
        {
          macros.apply(file, "ParserListVariableDeclaration", new String[] {"" + position});
        }
        else if(type_name.equals("null"))
        {
          // No intermediate variable needed for null arguments
        }
        else
        {
          macros.apply(file, "ParserSimpleVariableDeclaration", new String[] {type_name, type_name.toLowerCase(), "" + position});
        }
      }
      catch(IOException e)
      {
        throw new RuntimeException("An error occured while writing to " +
                                   new File(pkgDir, "Parser.java").getAbsolutePath(), e);
      }
    }
  }

  @Override
  public void outAAltTransform(AAltTransform node)
  {
    String type_name;
    int position;

    try
    {
      for(PTerm term : node.getTerms())
      {
        if(simpleTermTransformMap.get(term) != null)
        {
          type_name = simpleTermTransformMap.get(term);
        }
        else
        {
          type_name = CG.getAltTransformElemTypes().get(term);
        }

        position = CG.getTermNumbers().get(term);

        if(type_name.startsWith("L"))
        {
          type_name = "list";
        }
        else if(type_name.equals("null"))
        {
          type_name = "null";
        }
        else
        {
          type_name = type_name.toLowerCase();
        }
        macros.apply(file, "ParserNewBodyListAdd", new String[] {type_name, "" + position});

      }
      if (popCount > 0) {
        // The nodeArrayList variables are numbered starting at 1, so the first popped variable has the number popCount and not popCount-1.
        macros.apply(file, "ParserNewCheck", new String[] {String.valueOf(popCount)});
      }
      macros.apply(file, "ParserNewTail");
    }
    catch(IOException e)
    {
      throw new RuntimeException("An error occured while writing to " +
                                 new File(pkgDir, "Parser.java").getAbsolutePath(), e);
    }
  }

  public void inAParams(List<PTerm> list_param)
  {
    String type_name;
    int position;

    for(PTerm term : list_param)
    {
      if(simpleTermTransformMap.get(term) != null)
      {
        type_name = simpleTermTransformMap.get(term);
      }
      else
      {
        type_name = CG.getAltTransformElemTypes().get(term);
      }
      position = CG.getTermNumbers().get(term);

      try
      {
        if(type_name.startsWith("L"))
        {
          macros.apply(file, "ParserListVariableDeclaration", new String[] {"" + position});
        }
        else if(type_name.equals("null"))
        {
          // No intermediate variable needed for null arguments
        }
        else
        {
          macros.apply(file, "ParserSimpleVariableDeclaration", new String[] {type_name, type_name.toLowerCase(), "" + position});
        }
      }
      catch(IOException e)
      {
        throw new RuntimeException("An error occured while writing to " +
                                   new File(pkgDir, "Parser.java").getAbsolutePath(), e);
      }
    }
  }

  @Override
  public void inASimpleTerm(ASimpleTerm node)
  {
    try
    {
      String type_name;
      if(simpleTermTransformMap.get(node) != null)
      {
        type_name = simpleTermTransformMap.get(node);
      }
      else
      {
        type_name = CG.getAltTransformElemTypes().get(node);
      }
      int position = CG.getTermNumbers().get(node).intValue();
      String termKey = currentAlt+"."+node.getId().getText();
      int elemPosition = CTP.elems_position.get(termKey);
      int positionMap = 0;

      if(node.getSimpleTermTail() != null )
      {
        // FIXME This condition is always false, because listSimpleTermTransform contains ASimpleTerm/ASimpleListTerm and not String! Is it safe to remove this?
        if( !listSimpleTermTransform.contains(node.getId().getText() ) )
        {
          String type = CTP.positionsMap.get( realcurrentAlt+"."+node.getId().getText() );
          List<String> list = transformIds.getProdTransformIds().prod_transforms.get(type);
          if( list.indexOf( node.getSimpleTermTail().getText() ) >= 0 )
          {
            positionMap = list.indexOf( node.getSimpleTermTail().getText() );
          }
        }

        if(simpleTermOrsimpleListTermTypes.get(node) != null)
        {
          String type = simpleTermOrsimpleListTermTypes.get(node);
          List<String> list = transformIds.getProdTransformIds().prod_transforms.get(type);
          if( list.indexOf( node.getSimpleTermTail().getText() ) >= 0 )
          {
            positionMap = list.indexOf( node.getSimpleTermTail().getText() );
          }
        }
      }

      String type;
      if(type_name.startsWith("L"))
      {
        type_name = "list";
        type = "LinkedList";
      }
      else if(type_name.equals("null"))
      {
        type_name = "null";
        type = "Object";
      }
      else
      {
        type = type_name;
      }

      macros.apply(file, "ParserSimpleTerm", new String[]
                   {
                     type_name.toLowerCase(), ""+position,
                     type, ""+elemPosition, ""+positionMap
                   }
                  );
    }
    catch(IOException e)
    {
      throw new RuntimeException("An error occured while writing to " +
                                 new File(pkgDir, "Parser.java").getAbsolutePath(), e);
    }
  }

  @Override
  public void inASimpleListTerm(ASimpleListTerm node)
  {
    try
    {
      String type_name;
      if(simpleTermTransformMap.get(node) != null)
      {
        type_name = simpleTermTransformMap.get(node);
      }
      else
      {
        type_name = CG.getAltTransformElemTypes().get(node);
      }

      String termKey = currentAlt+"."+node.getId().getText();
      int position = CG.getTermNumbers().get(node);

      int elemPosition = CTP.elems_position.get(termKey);

      int positionMap = 0;

      if(node.getSimpleTermTail() != null )
      {
        // FIXME This condition is always false, because listSimpleTermTransform contains ASimpleTerm/ASimpleListTerm and not String! Is it safe to remove this?
        if( !listSimpleTermTransform.contains(node.getId().getText()) )
        {
          String type = CTP.positionsMap.get( realcurrentAlt+"."+node.getId().getText() );
          List<String> list = transformIds.getProdTransformIds().prod_transforms.get(type);
          if( list.indexOf( node.getSimpleTermTail().getText() ) >= 0 )
          {
            positionMap = list.indexOf( node.getSimpleTermTail().getText() );
          }
        }

        if(simpleTermOrsimpleListTermTypes.get(node) != null)
        {
          String type = simpleTermOrsimpleListTermTypes.get(node);
          List<String> list = transformIds.getProdTransformIds().prod_transforms.get(type);
          if( list.indexOf( node.getSimpleTermTail().getText() ) >= 0 )
          {
            positionMap = list.indexOf( node.getSimpleTermTail().getText() );
          }
        }
      }

      String type;
      if(type_name.startsWith("L"))
      {
        type_name = "list";
        type = "LinkedList";
      }
      else if(type_name.equals("null"))
      {
        type_name = "null";
        type = "Object";
      }
      else
      {
        type = type_name;
      }

      macros.apply(file, "ParserSimpleTerm", new String[]
                   {
                     type_name.toLowerCase(), ""+position,
                     type, ""+elemPosition, ""+positionMap
                   }
                  );
    }
    catch(IOException e)
    {
      throw new RuntimeException("An error occured while writing to " +
                                 new File(pkgDir, "Parser.java").getAbsolutePath(), e);
    }
  }

  @Override
  public void inANewTerm(ANewTerm node)
  {
    try
    {
      macros.apply(file, "ParserBraceOpening");
    }
    catch(IOException e)
    {
      throw new RuntimeException("An error occured while writing to " +
                                 new File(pkgDir, "Parser.java").getAbsolutePath(), e);
    }
    inAParams(node.getParams());
  }

  @Override
  public void inANewListTerm(ANewListTerm node)
  {
    try
    {
      macros.apply(file, "ParserBraceOpening");
    }
    catch(IOException e)
    {
      throw new RuntimeException("An error occured while writing to " +
                                 new File(pkgDir, "Parser.java").getAbsolutePath(), e);
    }
    inAParams(node.getParams());
  }

  @Override
  public void inAListTerm(AListTerm node)
  {
    try
    {
      macros.apply(file, "ParserBraceOpening");
      for(PListTerm listTerm : node.getListTerms())
      {
        String type_name;
        if(simpleTermTransformMap.get(listTerm) != null)
        {
          type_name = simpleTermTransformMap.get(listTerm);
        }
        else
        {
          type_name = CG.getAltTransformElemTypes().get(listTerm);
        }
        int position = CG.getTermNumbers().get(listTerm);

        if(type_name.startsWith("L"))
        {
          macros.apply(file, "ParserListVariableDeclaration", new String[] {"" + position});
        }
        else if(type_name.equals("null"))
        {
          // No intermediate variable needed for null arguments
        }
        else
        {
          macros.apply(file, "ParserSimpleVariableDeclaration", new String[] {type_name, type_name.toLowerCase(), "" + position});
        }
      }
    }
    catch(IOException e)
    {
      throw new RuntimeException("An error occured while writing to " +
                                 new File(pkgDir, "Parser.java").getAbsolutePath(), e);
    }
  }

  @Override
  public void outAListTerm(AListTerm node)
  {
    try
    {
      int listPosition = CG.getTermNumbers().get(node);

      for(PListTerm listTerm : node.getListTerms())
      {
        String type_name;
        if(simpleTermTransformMap.get(listTerm) != null)
        {
          type_name = simpleTermTransformMap.get(listTerm);
        }
        else
        {
          type_name = CG.getAltTransformElemTypes().get(listTerm);
        }
        int position = CG.getTermNumbers().get(listTerm);

        if(!type_name.equals("null"))
        {
          if(type_name.startsWith("L"))
          {
            macros.apply(file, "ParserTypedLinkedListAddAll", new String[] {"list", ""+listPosition, "list", ""+ position});
          }
          else
          {
            macros.apply(file, "ParserTypedLinkedListAdd", new String[] {"list", ""+listPosition, type_name.toLowerCase(), ""+ position});
          }
        }
      }
      macros.apply(file, "ParserBraceClosing");
    }
    catch(IOException e)
    {
      throw new RuntimeException("An error occured while writing to " +
                                 new File(pkgDir, "Parser.java").getAbsolutePath(), e);
    }
  }

  @Override
  public void outANewTerm(ANewTerm node)
  {
    String type_name;
    if(simpleTermTransformMap.get(node) != null)
    {
      type_name = simpleTermTransformMap.get(node);
    }
    else
    {
      type_name = CG.getAltTransformElemTypes().get(node);
    }
    int position = CG.getTermNumbers().get(node);
    String newAltName = name((AProdName)node.getProdName());

    try
    {
      if(type_name.startsWith("L"))
      {
        type_name = "list";
      }
      else
      {
        type_name = type_name.toLowerCase();
      }
      macros.apply(file, "ParserNewBodyNew", new String[] {type_name, ""+position, newAltName});

      if(node.getParams().size() > 0)
      {
        String isNotTheFirstParam = "";

        for(PTerm term : node.getParams())
        {
          if(simpleTermTransformMap.get(term) != null)
          {
            type_name = simpleTermTransformMap.get(term);
          }
          else
          {
            type_name = CG.getAltTransformElemTypes().get(term);
          }
          position = CG.getTermNumbers().get(term);

          if(type_name.equals("null"))
          {
            macros.apply(file, "ParserNew&ListBodyParamsNull", new String[] {isNotTheFirstParam+"null"});
          }
          else
          {
            if(type_name.startsWith("L"))
            {
              type_name = "list";
            }
            else
            {
              type_name = type_name.toLowerCase();
            }
            macros.apply(file, "ParserNew&ListBodyParams", new String[] {isNotTheFirstParam+type_name, ""+position});
          }

          isNotTheFirstParam = ", ";
        }
      }
      macros.apply(file, "ParserNewBodyNewTail");
      macros.apply(file, "ParserBraceClosing");
    }
    catch(IOException e)
    {
      throw new RuntimeException("An error occured while writing to " +
                                 new File(pkgDir, "TokenIndex.java").getAbsolutePath(), e);
    }
  }

  @Override
  public void outANewListTerm(ANewListTerm node)
  {
    String type_name;
    if(simpleTermTransformMap.get(node) != null)
    {
      type_name = simpleTermTransformMap.get(node);
    }
    else
    {
      type_name = CG.getAltTransformElemTypes().get(node);
    }
    int position = CG.getTermNumbers().get(node);
    String newAltName = name((AProdName)node.getProdName());
    try
    {
      if(type_name.startsWith("L"))
      {
        type_name = "list";
      }
      else
      {
        type_name = type_name.toLowerCase();
      }
      macros.apply(file, "ParserNewBodyNew", new String[] {type_name, ""+position, newAltName});

      if(node.getParams().size() > 0)
      {
        String isNotTheFirstParam = "";

        for(PTerm term : node.getParams())
        {
          if(simpleTermTransformMap.get(term) != null)
          {
            type_name = simpleTermTransformMap.get(term);
          }
          else
          {
            type_name = CG.getAltTransformElemTypes().get(term);
          }
          position = CG.getTermNumbers().get(term);

          if(type_name.equals("null"))
          {
            macros.apply(file, "ParserNew&ListBodyParamsNull", new String[] {isNotTheFirstParam+"null"});
          }
          else
          {
            if(type_name.startsWith("L"))
            {
              type_name = "list";
            }
            else
            {
              type_name = type_name.toLowerCase();
            }
            macros.apply(file, "ParserNew&ListBodyParams", new String[] {isNotTheFirstParam+type_name, ""+position});
          }

          isNotTheFirstParam = ", ";
        }
      }
      macros.apply(file, "ParserNewBodyNewTail");
      macros.apply(file, "ParserBraceClosing");
    }
    catch(IOException e)
    {
      throw new RuntimeException("An error occured while writing to " +
                                 new File(pkgDir, "TokenIndex.java").getAbsolutePath(), e);
    }
  }

  public String name(AProdName node)
  {
    if(node.getProdNameTail() != null)
    {
      return "A" +
             ResolveIds.name(node.getProdNameTail().getText()) +
             ResolveIds.name(node.getId().getText());
    }
    return "A" + ResolveIds.name(node.getId().getText());
  }

}
